@@ -24,8 +24,8 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 #endif
 
 #include <libdbusmenu-gtk/menu.h>
-#include <dbus/dbus-glib.h>
 #include <glib.h>
+#include <gio/gio.h>
 
 #include "window-menus.h"
 #include "indicator-appmenu-marshal.h"
@@ -37,14 +37,11 @@ struct _WindowMenusPrivate {
 	guint windowid;
 	DbusmenuGtkClient * client;
 	DbusmenuMenuitem * root;
-	DBusGProxy * props;
+	GCancellable * props_cancel;
+	GDBusProxy * props;
 	GArray * entries;
 	gboolean error_state;
 	guint   retry_timer;
-	gint    retry_id;
-	gchar * retry_name;
-	GValue  retry_data;
-	guint   retry_timestamp;
 };
 
 typedef struct _WMEntry WMEntry;
@@ -53,6 +50,7 @@ struct _WMEntry {
 	gboolean disabled;
 	gboolean hidden;
 	DbusmenuMenuitem * mi;
+	WindowMenus * wm;
 };
 
 #define WINDOW_MENUS_GET_PRIVATE(o) \
@@ -63,9 +61,10 @@ struct _WMEntry {
 enum {
 	ENTRY_ADDED,
 	ENTRY_REMOVED,
-	DESTROY,
 	ERROR_STATE,
+	STATUS_CHANGED,
 	SHOW_MENU,
+	A11Y_UPDATE,
 	LAST_SIGNAL
 };
 
@@ -76,13 +75,16 @@ static guint signals[LAST_SIGNAL] = { 0 };
 static void window_menus_class_init (WindowMenusClass *klass);
 static void window_menus_init       (WindowMenus *self);
 static void window_menus_dispose    (GObject *object);
-static void window_menus_finalize   (GObject *object);
-static void properties_destroyed    (GObject * object, gpointer user_data);
 static void root_changed            (DbusmenuClient * client, DbusmenuMenuitem * new_root, gpointer user_data);
+static void event_status            (DbusmenuClient * client, DbusmenuMenuitem * mi, gchar * event, GVariant * evdata, guint timestamp, GError * error, gpointer user_data);
+static void item_activate           (DbusmenuClient * client, DbusmenuMenuitem * item, guint timestamp, gpointer user_data);
+static void status_changed          (DbusmenuClient * client, GParamSpec * pspec, gpointer user_data);
 static void menu_entry_added        (DbusmenuMenuitem * root, DbusmenuMenuitem * newentry, guint position, gpointer user_data);
 static void menu_entry_removed      (DbusmenuMenuitem * root, DbusmenuMenuitem * oldentry, gpointer user_data);
 static void menu_entry_realized     (DbusmenuMenuitem * newentry, gpointer user_data);
+static void menu_prop_changed       (DbusmenuMenuitem * item, const gchar * property, GVariant * value, gpointer user_data);
 static void menu_child_realized     (DbusmenuMenuitem * child, gpointer user_data);
+static void props_cb (GObject * object, GAsyncResult * res, gpointer user_data);
 
 G_DEFINE_TYPE (WindowMenus, window_menus, G_TYPE_OBJECT);
 
@@ -95,7 +97,6 @@ window_menus_class_init (WindowMenusClass *klass)
 	g_type_class_add_private (klass, sizeof (WindowMenusPrivate));
 
 	object_class->dispose = window_menus_dispose;
-	object_class->finalize = window_menus_finalize;
 
 	/* Signals */
 	signals[ENTRY_ADDED] =  g_signal_new(WINDOW_MENUS_SIGNAL_ENTRY_ADDED,
@@ -112,13 +113,6 @@ window_menus_class_init (WindowMenusClass *klass)
 	                                      NULL, NULL,
 	                                      g_cclosure_marshal_VOID__POINTER,
 	                                      G_TYPE_NONE, 1, G_TYPE_POINTER);
-	signals[DESTROY] =       g_signal_new(WINDOW_MENUS_SIGNAL_DESTROY,
-	                                      G_TYPE_FROM_CLASS(klass),
-	                                      G_SIGNAL_RUN_LAST,
-	                                      G_STRUCT_OFFSET (WindowMenusClass, destroy),
-	                                      NULL, NULL,
-	                                      g_cclosure_marshal_VOID__VOID,
-	                                      G_TYPE_NONE, 0, G_TYPE_NONE);
 	signals[ERROR_STATE] =   g_signal_new(WINDOW_MENUS_SIGNAL_ERROR_STATE,
 	                                      G_TYPE_FROM_CLASS(klass),
 	                                      G_SIGNAL_RUN_LAST,
@@ -126,6 +120,13 @@ window_menus_class_init (WindowMenusClass *klass)
 	                                      NULL, NULL,
 	                                      g_cclosure_marshal_VOID__BOOLEAN,
 	                                      G_TYPE_NONE, 1, G_TYPE_BOOLEAN, G_TYPE_NONE);
+	signals[STATUS_CHANGED] = g_signal_new(WINDOW_MENUS_SIGNAL_STATUS_CHANGED,
+	                                      G_TYPE_FROM_CLASS(klass),
+	                                      G_SIGNAL_RUN_LAST,
+	                                      G_STRUCT_OFFSET (WindowMenusClass, status_changed),
+	                                      NULL, NULL,
+	                                      g_cclosure_marshal_VOID__INT,
+	                                      G_TYPE_NONE, 1, G_TYPE_INT, G_TYPE_NONE);
 	signals[SHOW_MENU] =     g_signal_new(WINDOW_MENUS_SIGNAL_SHOW_MENU,
 	                                      G_TYPE_FROM_CLASS(klass),
 	                                      G_SIGNAL_RUN_LAST,
@@ -133,6 +134,13 @@ window_menus_class_init (WindowMenusClass *klass)
 	                                      NULL, NULL,
 	                                      _indicator_appmenu_marshal_VOID__POINTER_UINT,
 	                                      G_TYPE_NONE, 2, G_TYPE_POINTER, G_TYPE_UINT, G_TYPE_NONE);
+	signals[A11Y_UPDATE] =   g_signal_new(WINDOW_MENUS_SIGNAL_A11Y_UPDATE,
+	                                      G_TYPE_FROM_CLASS(klass),
+	                                      G_SIGNAL_RUN_LAST,
+	                                      G_STRUCT_OFFSET (WindowMenusClass, a11y_update),
+	                                      NULL, NULL,
+	                                      _indicator_appmenu_marshal_VOID__POINTER,
+	                                      G_TYPE_NONE, 1, G_TYPE_POINTER, G_TYPE_NONE);
 
 	return;
 }
@@ -144,6 +152,7 @@ window_menus_init (WindowMenus *self)
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(self);
 
 	priv->client = NULL;
+	priv->props_cancel = NULL;
 	priv->props = NULL;
 	priv->root = NULL;
 	priv->error_state = FALSE;
@@ -153,20 +162,83 @@ window_menus_init (WindowMenus *self)
 	return;
 }
 
+static void
+entry_free(IndicatorObjectEntry * entry)
+{
+	g_return_if_fail(entry != NULL);
+	WMEntry * wmentry = (WMEntry *)entry;
+
+	if (wmentry->mi != NULL) {
+		g_signal_handlers_disconnect_by_func(wmentry->mi, G_CALLBACK(menu_prop_changed), &wmentry->ioentry);
+		g_object_unref(G_OBJECT(wmentry->mi));
+		wmentry->mi = NULL;
+	}
+
+	if (entry->label != NULL) {
+		g_object_unref(entry->label);
+		entry->label = NULL;
+	}
+	if (entry->accessible_desc != NULL) {
+		entry->accessible_desc = NULL;
+	}
+	if (entry->image != NULL) {
+		g_object_unref(entry->image);
+		entry->image = NULL;
+	}
+	if (entry->menu != NULL) {
+		g_signal_handlers_disconnect_by_func(entry->menu, G_CALLBACK(gtk_widget_destroyed), &entry->menu);
+		g_object_unref(entry->menu);
+		entry->menu = NULL;
+	}
+
+	g_free(entry);
+}
+
+static void
+free_entries(GObject *object, gboolean should_signal)
+{
+	g_return_if_fail(IS_WINDOW_MENUS(object));
+
+	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(object);
+
+	if (priv->entries != NULL) {
+		int i;
+		for (i = 0; i < priv->entries->len; i++) {
+			IndicatorObjectEntry * entry;
+			entry = g_array_index(priv->entries, IndicatorObjectEntry *, i);
+			g_array_remove_index(priv->entries, i);
+			if (should_signal) {			
+				g_signal_emit(object, signals[ENTRY_REMOVED], 0, entry, TRUE);
+			}
+			entry_free(entry);
+		}
+	}
+}
+
 /* Destroy objects */
 static void
 window_menus_dispose (GObject *object)
 {
-	g_signal_emit(object, signals[DESTROY], 0, TRUE);
-
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(object);
 
+	free_entries(object, FALSE);
+
+	if (priv->entries != NULL) {
+		g_array_free(priv->entries, TRUE);
+		priv->entries = NULL;
+	}
+
 	if (priv->root != NULL) {
-		g_object_unref(G_OBJECT(priv->root));
-		priv->root = NULL;
+		root_changed(DBUSMENU_CLIENT(priv->client), NULL, object);
+		g_warn_if_fail(priv->root == NULL);
 	}
 
 	if (priv->client != NULL) {
+		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(root_changed),    object);
+		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(event_status),    object);
+		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(item_activate),   object);
+		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->client), G_CALLBACK(status_changed),  object);
+
 		g_object_unref(G_OBJECT(priv->client));
 		priv->client = NULL;
 	}
@@ -176,48 +248,21 @@ window_menus_dispose (GObject *object)
 		priv->props = NULL;
 	}
 
+	if (priv->props_cancel != NULL) {
+		g_cancellable_cancel(priv->props_cancel);
+		g_object_unref(priv->props_cancel);
+		priv->props_cancel = NULL;
+	}
+
 	if (priv->retry_timer != 0) {
 		g_source_remove(priv->retry_timer);
 		priv->retry_timer = 0;
-		g_value_unset(&priv->retry_data);
-		g_value_reset(&priv->retry_data);
-		g_free(priv->retry_name);
-		priv->retry_name = NULL;
 	}
 
 	G_OBJECT_CLASS (window_menus_parent_class)->dispose (object);
 	return;
 }
 
-/* Free memory */
-static void
-window_menus_finalize (GObject *object)
-{
-	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(object);
-
-	g_debug("Window Menus Object finalizing for: %d", priv->windowid);
-
-	if (priv->entries != NULL) {
-		int i;
-		for (i = 0; i < priv->entries->len; i++) {
-			IndicatorObjectEntry * entry;
-			entry = g_array_index(priv->entries, IndicatorObjectEntry *, i);
-			
-			if (entry->label != NULL) {
-				g_object_unref(entry->label);
-			}
-			if (entry->menu != NULL) {
-				g_object_unref(entry->menu);
-			}
-		}
-		g_array_free(priv->entries, TRUE);
-		priv->entries = NULL;
-	}
-
-	G_OBJECT_CLASS (window_menus_parent_class)->finalize (object);
-	return;
-}
-
 /* Retry the event sending to the server to see if we can get things
    working again. */
 static gboolean
@@ -227,20 +272,19 @@ retry_event (gpointer user_data)
 	g_return_val_if_fail(IS_WINDOW_MENUS(user_data), FALSE);
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(user_data);
 
-	dbusmenu_client_send_event(DBUSMENU_CLIENT(priv->client), priv->retry_id, priv->retry_name, &priv->retry_data, priv->retry_timestamp);
+	dbusmenu_menuitem_handle_event(dbusmenu_client_get_root(DBUSMENU_CLIENT(priv->client)),
+	                               "x-appmenu-retry-ping",
+	                               NULL,
+	                               0);
 
 	priv->retry_timer = 0;
-	g_value_unset(&priv->retry_data);
-	g_value_reset(&priv->retry_data);
-	g_free(priv->retry_name);
-	priv->retry_name = NULL;
 
 	return FALSE;
 }
 
 /* Listen to whether our events are successfully sent */
 static void
-event_status (DbusmenuClient * client, DbusmenuMenuitem * mi, gchar * event, GValue * evdata, guint timestamp, GError * error, gpointer user_data)
+event_status (DbusmenuClient * client, DbusmenuMenuitem * mi, gchar * event, GVariant * evdata, guint timestamp, GError * error, gpointer user_data)
 {
 	g_return_if_fail(IS_WINDOW_MENUS(user_data));
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(user_data);
@@ -266,10 +310,6 @@ event_status (DbusmenuClient * client, DbusmenuMenuitem * mi, gchar * event, GVa
 		if (priv->retry_timer != 0) {
 			g_source_remove(priv->retry_timer);
 			priv->retry_timer = 0;
-			g_value_unset(&priv->retry_data);
-			g_value_reset(&priv->retry_data);
-			g_free(priv->retry_name);
-			priv->retry_name = NULL;
 		}
 
 		return;
@@ -294,17 +334,31 @@ event_status (DbusmenuClient * client, DbusmenuMenuitem * mi, gchar * event, GVa
 	if (priv->retry_timer == 0) {
 		g_debug("Setting up retry timer");
 		priv->retry_timer = g_timeout_add_seconds(1, retry_event, user_data);
-
-		priv->retry_id = dbusmenu_menuitem_get_id(mi);
-		priv->retry_name = g_strdup(event);
-		g_value_init(&priv->retry_data, G_VALUE_TYPE(evdata));
-		g_value_copy(evdata, &priv->retry_data);
-		priv->retry_timestamp = timestamp;
 	}
 
 	return;
 }
 
+static IndicatorObjectEntry *
+get_entry(WindowMenus *wm, DbusmenuMenuitem * item, guint *index)
+{
+	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(wm);
+
+	guint position = 0;
+	for (position = 0; position < priv->entries->len; ++position) {
+		WMEntry * entry = g_array_index(priv->entries, WMEntry *, position);
+		if (entry->mi == item) {
+			if (index != NULL) {
+				*index = position;
+			}
+			return &entry->ioentry;
+		}
+	}
+
+	/* Not found */
+	return NULL;
+}
+
 /* Called when a menu item wants to be displayed.  We need to see if
    it's one of our root items and pass it up if so. */
 static void
@@ -317,22 +371,34 @@ item_activate (DbusmenuClient * client, DbusmenuMenuitem * item, guint timestamp
 		return;
 	}
 
-	guint position = dbusmenu_menuitem_get_position(priv->root, item);
-	if (position == 0) {
-		/* Ugly, ugly hack.  I shouldn't have used guint in the function
-		   above, but now I have to do this.  Ew! */
-		GList * children = dbusmenu_menuitem_get_children(priv->root);
-		if (children == NULL || children->data != item) {
-			return;
-		}
+	IndicatorObjectEntry * entry = get_entry(WINDOW_MENUS(user_data), item, NULL);
+	if (entry == NULL) {
+		/* Not found */
+		return;
 	}
 
-	IndicatorObjectEntry * entry = g_array_index(priv->entries, IndicatorObjectEntry *, position);
 	g_signal_emit(G_OBJECT(user_data), signals[SHOW_MENU], 0, entry, timestamp, TRUE);
 
 	return;
 }
 
+/* Called when the client changes its status.  Used to show panel if requested.
+   (Say, by an Alt press.) */
+static void
+status_changed (DbusmenuClient * client, GParamSpec * pspec, gpointer user_data)
+{
+	g_signal_emit(G_OBJECT(user_data), signals[STATUS_CHANGED], 0, dbusmenu_client_get_status (client));
+}
+
+DbusmenuStatus
+window_menus_get_status (WindowMenus * wm)
+{
+	g_return_val_if_fail(IS_WINDOW_MENUS(wm), DBUSMENU_STATUS_NORMAL);
+	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(wm);
+
+	return dbusmenu_client_get_status (DBUSMENU_CLIENT (priv->client));
+}
+
 /* Build a new window menus object and attach to the signals to build
    up the representative menu. */
 WindowMenus *
@@ -344,34 +410,33 @@ window_menus_new (const guint windowid, const gchar * dbus_addr, const gchar * d
 	g_return_val_if_fail(dbus_addr != NULL, NULL);
 	g_return_val_if_fail(dbus_object != NULL, NULL);
 
-	DBusGConnection * session_bus = dbus_g_bus_get(DBUS_BUS_SESSION, NULL);
-	g_return_val_if_fail(session_bus != NULL, NULL);
-
 	WindowMenus * newmenu = WINDOW_MENUS(g_object_new(WINDOW_MENUS_TYPE, NULL));
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(newmenu);
 
 	priv->windowid = windowid;
 
-	priv->props = dbus_g_proxy_new_for_name_owner(session_bus,
-	                                              dbus_addr,
-	                                              dbus_object,
-	                                              DBUS_INTERFACE_PROPERTIES,
-	                                              NULL);
-	if (priv->props == NULL) {
-		g_warning("Unable to get property proxy on '%s' object '%s'", dbus_addr, dbus_object);
-		g_object_unref(newmenu);
-		return NULL;
-	}
-
-	g_signal_connect(G_OBJECT(priv->props), "destroy", G_CALLBACK(properties_destroyed), newmenu);
+	/* Build the service proxy */
+	priv->props_cancel = g_cancellable_new();
+	g_object_ref(newmenu); /* Take a ref for the async callback */
+	g_dbus_proxy_new_for_bus(G_BUS_TYPE_SESSION,
+	                         G_DBUS_PROXY_FLAGS_NONE,
+	                         NULL,
+	                         dbus_addr,
+	                         dbus_object,
+	                         "org.freedesktop.DBus.Properties",
+	                         priv->props_cancel,
+	                         props_cb,
+	                         newmenu);
 
 	priv->client = dbusmenu_gtkclient_new((gchar *)dbus_addr, (gchar *)dbus_object);
 	GtkAccelGroup * agroup = gtk_accel_group_new();
 	dbusmenu_gtkclient_set_accel_group(priv->client, agroup);
+	g_object_unref(agroup);
 
 	g_signal_connect(G_OBJECT(priv->client), DBUSMENU_GTKCLIENT_SIGNAL_ROOT_CHANGED, G_CALLBACK(root_changed),   newmenu);
 	g_signal_connect(G_OBJECT(priv->client), DBUSMENU_CLIENT_SIGNAL_EVENT_RESULT, G_CALLBACK(event_status), newmenu);
 	g_signal_connect(G_OBJECT(priv->client), DBUSMENU_CLIENT_SIGNAL_ITEM_ACTIVATE, G_CALLBACK(item_activate), newmenu);
+	g_signal_connect(G_OBJECT(priv->client), "notify::" DBUSMENU_CLIENT_PROP_STATUS, G_CALLBACK(status_changed), newmenu);
 
 	DbusmenuMenuitem * root = dbusmenu_client_get_root(DBUSMENU_CLIENT(priv->client));
 	if (root != NULL) {
@@ -381,18 +446,42 @@ window_menus_new (const guint windowid, const gchar * dbus_addr, const gchar * d
 	return newmenu;
 }
 
-/* Respond to the proxies getting destoryed.  I means that we need
-   to kill ourselves. */
+/* Callback from trying to create the proxy for the service, this
+   could include starting the service. */
 static void
-properties_destroyed (GObject * object, gpointer user_data)
+props_cb (GObject * object, GAsyncResult * res, gpointer user_data)
 {
-	WindowMenus * wm = WINDOW_MENUS(user_data);
-	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(wm);
+	GError * error = NULL;
+	GDBusProxy * proxy = g_dbus_proxy_new_for_bus_finish(res, &error);
 
-	priv->props = NULL;
-	g_debug("Properties destroyed for window: %d", priv->windowid);
+	if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
+		g_error_free (error);
+		return; // Must exit before accessing freed memory
+	}
+
+	WindowMenus * self = WINDOW_MENUS(user_data);
+	g_return_if_fail(self != NULL);
+
+	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(self);
+
+	if (priv->props_cancel != NULL) {
+		g_object_unref(priv->props_cancel);
+		priv->props_cancel = NULL;
+	}
+
+	if (error != NULL) {
+		g_warning("Could not grab DBus proxy for window %u: %s", priv->windowid, error->message);
+		g_error_free(error);
+		goto out;
+	}
+
+	/* Okay, we're good to grab the proxy at this point, we're
+	sure that it's ours. */
+	priv->props = proxy;
+
+out:
+	g_object_unref(G_OBJECT(self));
 
-	g_object_unref(G_OBJECT(wm));
 	return;
 }
 
@@ -448,6 +537,18 @@ new_root_helper (DbusmenuMenuitem * item, gpointer user_data)
 	return;
 }
 
+/* Remove the various signals that we attach to menuitems to
+   ensure they don't pop up later. */
+static void
+remove_menuitem_signals (DbusmenuMenuitem * mi, gpointer user_data)
+{
+	g_signal_handlers_disconnect_by_func(G_OBJECT(mi), G_CALLBACK(menu_entry_realized), user_data);
+	g_signal_handlers_disconnect_matched (mi, G_SIGNAL_MATCH_FUNC, 0, 0, 0, menu_child_realized, NULL);
+	g_signal_handlers_disconnect_matched (mi, G_SIGNAL_MATCH_FUNC, 0, 0, 0, menu_prop_changed, NULL);
+
+	return;
+}
+
 /* Respond to the root menu item on our client changing */
 static void
 root_changed (DbusmenuClient * client, DbusmenuMenuitem * new_root, gpointer user_data)
@@ -456,11 +557,11 @@ root_changed (DbusmenuClient * client, DbusmenuMenuitem * new_root, gpointer use
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(user_data);
 
 	/* Remove the old entries */
-	while (priv->entries->len != 0) {
-		menu_entry_removed(NULL, NULL, user_data);
-	}
+	free_entries(G_OBJECT(user_data), TRUE);
 
 	if (priv->root != NULL) {
+		dbusmenu_menuitem_foreach(priv->root, remove_menuitem_signals, client);
+
 		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->root), G_CALLBACK(menu_entry_added), user_data);
 		g_signal_handlers_disconnect_by_func(G_OBJECT(priv->root), G_CALLBACK(menu_entry_removed), user_data);
 		g_object_unref(priv->root);
@@ -494,6 +595,7 @@ root_changed (DbusmenuClient * client, DbusmenuMenuitem * new_root, gpointer use
 static void
 menu_entry_added (DbusmenuMenuitem * root, DbusmenuMenuitem * newentry, guint position, gpointer user_data)
 {
+	g_return_if_fail(IS_WINDOW_MENUS(user_data));
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(user_data);
 
 	g_signal_connect(G_OBJECT(newentry), DBUSMENU_MENUITEM_SIGNAL_REALIZED, G_CALLBACK(menu_entry_realized), user_data);
@@ -505,29 +607,54 @@ menu_entry_added (DbusmenuMenuitem * root, DbusmenuMenuitem * newentry, guint po
 	return;
 }
 
+/* A small clean up function to ensure that the data
+   gets free'd and the ref lost in all cases. */
+static void
+child_realized_data_cleanup (gpointer user_data, GClosure * closure)
+{
+	gpointer * data = (gpointer *)user_data;
+	g_object_unref(data[1]);
+	g_free(user_data);
+	return;
+}
+
 /* React to the menuitem when we know that it's got all the data
    that we really need. */
 static void
 menu_entry_realized (DbusmenuMenuitem * newentry, gpointer user_data)
 {
+	g_return_if_fail(IS_WINDOW_MENUS(user_data));
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(user_data);
+
+	/* Remove handler once we're called, this shouldn't be an issue for
+	   the realized signal, but we're using it for child added as well
+	   and in that case it could happen more than once. */
+	g_signal_handlers_disconnect_by_func(G_OBJECT(newentry), G_CALLBACK(menu_entry_realized), user_data);
+
 	GtkMenu * menu = dbusmenu_gtkclient_menuitem_get_submenu(priv->client, newentry);
 
+	/* Check to see if we have any children, if we don't let's see if
+	   we can scare some up for fun. */
+	GList * children = dbusmenu_menuitem_get_children(newentry);
+	if (children == NULL && g_strcmp0(DBUSMENU_MENUITEM_CHILD_DISPLAY_SUBMENU, dbusmenu_menuitem_property_get(newentry, DBUSMENU_MENUITEM_PROP_CHILD_DISPLAY)) == 0) {
+		dbusmenu_menuitem_send_about_to_show(newentry, NULL, NULL);
+	}
+
 	if (menu == NULL) {
-		GList * children = dbusmenu_menuitem_get_children(newentry);
 		if (children != NULL) {
 			gpointer * data = g_new(gpointer, 2);
 			data[0] = user_data;
-			data[1] = newentry;
+			data[1] = g_object_ref(newentry);
 
-			g_signal_connect(G_OBJECT(children->data), DBUSMENU_MENUITEM_SIGNAL_REALIZED, G_CALLBACK(menu_child_realized), data);
+			g_signal_connect_data(G_OBJECT(children->data), DBUSMENU_MENUITEM_SIGNAL_REALIZED, G_CALLBACK(menu_child_realized), data, child_realized_data_cleanup, 0);
 		} else {
-			g_warning("Entry has no children!");
+			g_signal_connect(G_OBJECT(newentry), DBUSMENU_MENUITEM_SIGNAL_CHILD_ADDED, G_CALLBACK(menu_entry_realized), user_data);
 		}
 	} else {
 		gpointer * data = g_new(gpointer, 2);
 		data[0] = user_data;
-		data[1] = newentry;
+		data[1] = g_object_ref(newentry);
+		/* child_realized does an unref */
 
 		menu_child_realized(NULL, data);
 	}
@@ -538,13 +665,13 @@ menu_entry_realized (DbusmenuMenuitem * newentry, gpointer user_data)
 /* Respond to properties changing on the menu item so that we can
    properly hide and show them. */
 static void
-menu_prop_changed (DbusmenuMenuitem * item, const gchar * property, const GValue * value, gpointer user_data)
+menu_prop_changed (DbusmenuMenuitem * item, const gchar * property, GVariant * value, gpointer user_data)
 {
 	IndicatorObjectEntry * entry = (IndicatorObjectEntry *)user_data;
 	WMEntry * wmentry = (WMEntry *)user_data;
 
 	if (!g_strcmp0(property, DBUSMENU_MENUITEM_PROP_VISIBLE)) {
-		if (g_value_get_boolean(value)) {
+		if (g_variant_get_boolean(value)) {
 			gtk_widget_show(GTK_WIDGET(entry->label));
 			wmentry->hidden = FALSE;
 		} else {
@@ -552,10 +679,15 @@ menu_prop_changed (DbusmenuMenuitem * item, const gchar * property, const GValue
 			wmentry->hidden = TRUE;
 		}
 	} else if (!g_strcmp0(property, DBUSMENU_MENUITEM_PROP_ENABLED)) {
-		gtk_widget_set_sensitive(GTK_WIDGET(entry->label), g_value_get_boolean(value));
-		wmentry->disabled = !g_value_get_boolean(value);
+		gtk_widget_set_sensitive(GTK_WIDGET(entry->label), g_variant_get_boolean(value));
+		wmentry->disabled = !g_variant_get_boolean(value);
 	} else if (!g_strcmp0(property, DBUSMENU_MENUITEM_PROP_LABEL)) {
-		gtk_label_set_text_with_mnemonic(entry->label, g_value_get_string(value));
+		gtk_label_set_text_with_mnemonic(entry->label, g_variant_get_string(value, NULL));
+		entry->accessible_desc = g_variant_get_string(value, NULL);
+
+		if (wmentry->wm != NULL) {
+			g_signal_emit(G_OBJECT(wmentry->wm), A11Y_UPDATE, 0, entry, TRUE);
+		}
 	}
 
 	return;
@@ -565,16 +697,29 @@ menu_prop_changed (DbusmenuMenuitem * item, const gchar * property, const GValue
 static void
 menu_child_realized (DbusmenuMenuitem * child, gpointer user_data)
 {
-	DbusmenuMenuitem * newentry = (DbusmenuMenuitem *)(((gpointer *)user_data)[1]);
+	/* Grab our values out to stack variables */
+	DbusmenuMenuitem * newentry = DBUSMENU_MENUITEM(((gpointer *)user_data)[1]);
+	WindowMenus * wm = WINDOW_MENUS(((gpointer *)user_data)[0]);
+
+	g_return_if_fail(newentry != NULL);
+	g_return_if_fail(wm != NULL);
+
+	/* Disconnection below will drop the ref for this signal
+	   handler, let's make sure that's not a problem */
+	g_object_ref(G_OBJECT(newentry));
 
 	/* Only care about the first */
+	/* This will cause the cleanup function attached to the signal
+	   handler to be run. */
 	g_signal_handlers_disconnect_by_func(G_OBJECT(child), menu_child_realized, user_data);
 
-	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE((((gpointer *)user_data)[0]));
+	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(wm);
 	WMEntry * wmentry = g_new0(WMEntry, 1);
+	wmentry->wm = wm;
 	IndicatorObjectEntry * entry = &wmentry->ioentry;
 
 	wmentry->mi = newentry;
+	g_object_ref(G_OBJECT(wmentry->mi));
 
 	entry->label = GTK_LABEL(gtk_label_new_with_mnemonic(dbusmenu_menuitem_property_get(newentry, DBUSMENU_MENUITEM_PROP_LABEL)));
 
@@ -582,6 +727,8 @@ menu_child_realized (DbusmenuMenuitem * child, gpointer user_data)
 		g_object_ref(entry->label);
 	}
 
+	entry->accessible_desc = dbusmenu_menuitem_property_get(newentry, DBUSMENU_MENUITEM_PROP_LABEL);
+
 	entry->menu = dbusmenu_gtkclient_menuitem_get_submenu(priv->client, newentry);
 
 	if (entry->menu == NULL) {
@@ -589,11 +736,12 @@ menu_child_realized (DbusmenuMenuitem * child, gpointer user_data)
 	} else {
 		g_object_ref(entry->menu);
 		gtk_menu_detach(entry->menu);
+		g_signal_connect(entry->menu, "destroy", G_CALLBACK(gtk_widget_destroyed), &entry->menu);
 	}
 
 	g_signal_connect(G_OBJECT(newentry), DBUSMENU_MENUITEM_SIGNAL_PROPERTY_CHANGED, G_CALLBACK(menu_prop_changed), entry);
 
-	if (dbusmenu_menuitem_property_get_value(newentry, DBUSMENU_MENUITEM_PROP_VISIBLE) != NULL
+	if (dbusmenu_menuitem_property_get_variant(newentry, DBUSMENU_MENUITEM_PROP_VISIBLE) != NULL
 		&& dbusmenu_menuitem_property_get_bool(newentry, DBUSMENU_MENUITEM_PROP_VISIBLE) == FALSE) {
 		gtk_widget_hide(GTK_WIDGET(entry->label));
 		wmentry->hidden = TRUE;
@@ -602,7 +750,7 @@ menu_child_realized (DbusmenuMenuitem * child, gpointer user_data)
 		wmentry->hidden = FALSE;
 	}
 
-	if (dbusmenu_menuitem_property_get_value (newentry, DBUSMENU_MENUITEM_PROP_ENABLED) != NULL) {
+	if (dbusmenu_menuitem_property_get_variant (newentry, DBUSMENU_MENUITEM_PROP_ENABLED) != NULL) {
 		gboolean sensitive = dbusmenu_menuitem_property_get_bool(newentry, DBUSMENU_MENUITEM_PROP_ENABLED);
 		gtk_widget_set_sensitive(GTK_WIDGET(entry->label), sensitive);
 		wmentry->disabled = !sensitive;
@@ -610,9 +758,9 @@ menu_child_realized (DbusmenuMenuitem * child, gpointer user_data)
 
 	g_array_append_val(priv->entries, wmentry);
 
-	g_signal_emit(G_OBJECT((((gpointer *)user_data)[0])), signals[ENTRY_ADDED], 0, entry, TRUE);
+	g_signal_emit(G_OBJECT(wm), signals[ENTRY_ADDED], 0, entry, TRUE);
 
-	g_free(user_data);
+	g_object_unref(newentry);
 
 	return;
 }
@@ -622,19 +770,26 @@ static void
 menu_entry_removed (DbusmenuMenuitem * root, DbusmenuMenuitem * oldentry, gpointer user_data)
 {
 	g_return_if_fail(IS_WINDOW_MENUS(user_data));
+	g_return_if_fail(DBUSMENU_IS_MENUITEM(oldentry));
 	WindowMenusPrivate * priv = WINDOW_MENUS_GET_PRIVATE(user_data);
 
 	if (priv->entries == NULL || priv->entries->len == 0) {
 		return;
 	}
-	
-	/* TODO: find the menuitem */
-	IndicatorObjectEntry * entry = g_array_index(priv->entries, IndicatorObjectEntry *, priv->entries->len - 1);
-	g_array_remove_index(priv->entries, priv->entries->len - 1);
 
-	g_signal_emit(G_OBJECT(user_data), signals[ENTRY_REMOVED], 0, entry, TRUE);
+	guint position;
+	IndicatorObjectEntry * entry = get_entry(WINDOW_MENUS(user_data), oldentry, &position);
 
-	g_free(entry);
+	if (entry != NULL) {
+		g_array_remove_index(priv->entries, position);
+		g_signal_emit(G_OBJECT(user_data), signals[ENTRY_REMOVED], 0, entry, TRUE);
+		entry_free(entry);
+	} else {
+		/* We've been called before menu_child_realized fired,
+		 * so there isn't a WMEntry yet. Don't be going ahead
+		 * and creating one if this menu item continues to live! */
+		g_signal_handlers_disconnect_by_func(G_OBJECT(oldentry), G_CALLBACK(menu_entry_realized), user_data);
+	}
 
 	return;
 }
